/********************************************************************************
*                                                                               *
*                     T h e   A d i e   T e x t   E d i t o r                   *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998,2003 by Jeroen van der Zijp.   All Rights Reserved.        *
*********************************************************************************
* This program is free software; you can redistribute it and/or modify          *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation; either version 2 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* This program is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with this program; if not, write to the Free Software                   *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
*********************************************************************************
* $Id: Adie.cpp,v 1.85 2003/02/10 06:09:07 fox Exp $                            *
********************************************************************************/
#include "fx.h"
#include "fxkeys.h"
#include "FXRex.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <signal.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <ctype.h>
#include "HelpWindow.h"
#include "Preferences.h"
#include "Commands.h"
#include "Hilite.h"
#include "TextWindow.h"
#include "Adie.h"
#include "icons.h"

/*
  Notes:
  - Manage TextWindows (there are more than one).
  - Open new window or open existing file in window; make sure
    no window already shows file to be opened (check based on abs pathname).
  - Common styles kept here.
  - One single collection of icons.
*/


/*******************************************************************************/

// Map
FXDEFMAP(Adie) AdieMap[]={
  FXMAPFUNC(SEL_SIGNAL,Adie::ID_QUIT,Adie::onCmdQuit),
  FXMAPFUNC(SEL_COMMAND,Adie::ID_QUIT,Adie::onCmdQuit),
  FXMAPFUNC(SEL_COMMAND,Adie::ID_NEW,Adie::onCmdNew),
  FXMAPFUNC(SEL_COMMAND,Adie::ID_OPEN,Adie::onCmdOpen),
  FXMAPFUNC(SEL_COMMAND,Adie::ID_OPEN_RECENT,Adie::onCmdOpenRecent),
  FXMAPFUNC(SEL_COMMAND,Adie::ID_OPEN_SELECTED,Adie::onCmdOpenSelected),
  FXMAPFUNC(SEL_CLOSE,Adie::ID_CLOSE,Adie::onCmdClose),
  FXMAPFUNC(SEL_UPDATE,Adie::ID_CLOSE,Adie::onUpdTitle),
  FXMAPFUNCS(SEL_UPDATE,Adie::ID_WINDOW_1,Adie::ID_WINDOW_10,Adie::onUpdWindowSelect),
  FXMAPFUNCS(SEL_COMMAND,Adie::ID_WINDOW_1,Adie::ID_WINDOW_10,Adie::onCmdWindowSelect),
  };


// Object implementation
FXIMPLEMENT(Adie,FXApp,AdieMap,ARRAYNUMBER(AdieMap))


// Name of untitled buffer
const FXchar Adie::untitled[]="untitled";


// Make some windows
Adie::Adie(const FXString& name):FXApp(name,FXString::null){

  // Make some icons; these are shared between all text windows
  bigicon=new FXGIFIcon(this,big_gif);
  smallicon=new FXGIFIcon(this,small_gif);
  newicon=new FXGIFIcon(this,new_gif,0,IMAGE_ALPHAGUESS);
  reloadicon=new FXGIFIcon(this,reload_gif);
  openicon=new FXGIFIcon(this,open_gif);
  saveicon=new FXGIFIcon(this,save_gif);
  saveasicon=new FXGIFIcon(this,saveas_gif,0,IMAGE_ALPHAGUESS);
  printicon=new FXGIFIcon(this,print_gif);
  cuticon=new FXGIFIcon(this,cut_gif);
  copyicon=new FXGIFIcon(this,copy_gif);
  pasteicon=new FXGIFIcon(this,paste_gif);
  deleteicon=new FXGIFIcon(this,delete_gif);
  undoicon=new FXGIFIcon(this,undo_gif);
  redoicon=new FXGIFIcon(this,redo_gif);
  fontsicon=new FXGIFIcon(this,fonts_gif);
  helpicon=new FXGIFIcon(this,help_gif);
  quiticon=new FXGIFIcon(this,quit_gif);
  shiftlefticon=new FXGIFIcon(this,shiftleft_gif);
  shiftrighticon=new FXGIFIcon(this,shiftright_gif);

  searchicon=new FXGIFIcon(this,search_gif,0,IMAGE_ALPHAGUESS);
  searchnexticon=new FXGIFIcon(this,searchnext_gif,0,IMAGE_ALPHAGUESS);
  searchprevicon=new FXGIFIcon(this,searchprev_gif,0,IMAGE_ALPHAGUESS);
  bookseticon=new FXGIFIcon(this,bookset_gif);
  booknexticon=new FXGIFIcon(this,booknext_gif);
  bookprevicon=new FXGIFIcon(this,bookprev_gif);
  bookdelicon=new FXGIFIcon(this,bookdel_gif);

  // If interrupt happens, quit gracefully; we may want to
  // save edit buffer contents w/o asking if display gets
  // disconnected or if hangup signal is received.
  addSignal(SIGINT,this,ID_QUIT);
#ifndef WIN32
  addSignal(SIGQUIT,this,ID_QUIT);
  addSignal(SIGHUP,this,ID_QUIT);
  addSignal(SIGPIPE,this,ID_QUIT);
#endif

  // Recent files
  mrufiles.setTarget(this);
  mrufiles.setSelector(ID_OPEN_RECENT);

  // File associations
  associations=new FXFileDict(this);
  }


// Initialize application
void Adie::init(int& argc,char** argv,FXbool connect){

  // After init, the registry has been loaded
  FXApp::init(argc,argv,connect);

  // Now we know the icon search path
  associations->setIconPath(reg().readStringEntry("SETTINGS","iconpath",FXIconDict::defaultIconPath));
  }


// Determine unique name
FXString Adie::uniqueName() const {
  FXString name=Adie::untitled;
  for(int i=1; i<2147483647; i++){
    if(!findWindow(name)) break;
    name.format("%s%d",Adie::untitled,i);
    }
  return name;
  }


// Find as yet untitled window
TextWindow *Adie::findUntitled() const {
  for(FXint w=0; w<windowlist.no(); w++){
    if(!windowlist[w]->isFilenameSet() && !windowlist[w]->isModified()) return windowlist[w];
    }
  return NULL;
  }


// Find window, if any, currently editing the given file
TextWindow *Adie::findWindow(const FXString& file) const {
  for(FXint w=0; w<windowlist.no(); w++){
    if(windowlist[w]->getFilename()==file) return windowlist[w];
    }
  return NULL;
  }


// Get new window; should search for currently unused window first
TextWindow *Adie::newWindow(const FXString& file){
  TextWindow *window;

  window=findUntitled();
  if(window){
    window->raise();
    window->setFocus();
    return window;
    }

  // Make new window
  window=new TextWindow(this,file);

  // Add to list
  windowlist.append(window);

  // Create it and raise to top
  window->create();
  window->raise();
  window->setFocus();

  return window;
  }


// Open existing file; first try find if its already open
TextWindow *Adie::openWindow(const FXString& file){
  TextWindow *window;

  // Check if already being edited by another window
  window=findWindow(file);
  if(window){
    window->raise();
    window->setFocus();
    return window;
    }

  // Make new window
  window=newWindow(file);

  // Try open the file
  window->loadFile(file);
  window->readBookmarks(file);
  window->readView(file);

  return window;
  }


// Close window
void Adie::closeWindow(TextWindow *window){
  if(window->saveChanges()){
    windowlist.remove(window);
    }
  }


// Close all windows
void Adie::closeAllWindows(){
  for(int i=0; i<windowlist.no(); i++){
    delete windowlist[i];
    }
  windowlist.clear();
  }


// New
long Adie::onCmdNew(FXObject*,FXSelector,void*){
  newWindow(uniqueName());
  return 1;
  }


// Open
long Adie::onCmdOpen(FXObject*,FXSelector,void*){
  FXString file;
  FXFileDialog opendialog(this,"Open Document");
  opendialog.setSelectMode(SELECTFILE_EXISTING);
//  opendialog.setPatternList(getPatterns());
//  opendialog.setCurrentPattern(getCurrentPattern());
//  opendialog.setFilename(filename);
  if(opendialog.execute()){
//    setCurrentPattern(opendialog.getCurrentPattern());
    file=opendialog.getFilename();
    openWindow(opendialog.getFilename());
    }
  return 1;
  }


// Open recent file
long Adie::onCmdOpenRecent(FXObject*,FXSelector,void* ptr){
  openWindow((const char*)ptr);
  return 1;
  }
  
  
// Open Selected
long Adie::onCmdOpenSelected(FXObject*,FXSelector,void*){
/*
  FXchar name[1024],*data;
  FXuint len;
  FXint pos=0,lineno=0;

  // Get selection
  if(windowlist[0]->getDNDData(FROM_SELECTION,FXWindow::stringType,(FXuchar*&)data,len)){
    FXString string(data,len);
    FXFREE(&data);

    // Its too big, most likely not a file name
    if(len<1024){
      FXString file=FXString::null;
      FXString dir=FXFile::getCurrentDirectory();

      // Base off currently loaded file
      if(!filename.empty()) dir=FXFile::directory(filename);

      // Strip leading/trailing space
      string.trim();

      // Extract name from #include syntax
      if(sscanf(string.text(),"#include \"%[^\"]\"",name)==1){
        file=FXFile::absolute(dir,name);
        if(!FXFile::exists(file)){
          file=FXFile::search(searchpath,name);
          }
        }
      else if(sscanf(string.text(),"#include <%[^>]>",name)==1){
        file=FXFile::absolute(dir,name);
        if(!FXFile::exists(file)){
          file=FXFile::search(searchpath,name);
          }
        }

      // Compiler output in the form filename:lineno: Error message
      else if(sscanf(string.text(),"%[^:]:%d:",name,&lineno)==2){
        file=FXFile::absolute(dir,name);
        if(!FXFile::exists(file)){
          file=FXFile::absolute(dir,string);
          }
        }

      // Compiler output in the form filename: Other stuff
      else if(sscanf(string.text(),"%[^:]:",name)==1){
        file=FXFile::absolute(dir,name);
        if(!FXFile::exists(file)){
          file=FXFile::absolute(dir,string);
          }
        }

      // Try whole selection
      else{
        file=FXFile::absolute(dir,string);
        }

      // Not a file name
      if(FXFile::exists(file)){

        // Different from current file?
        if(file!=filename){

          // Save old file
     //     if(!saveChanges()) return 1;

          // Open it
      //    loadFile(file);
          }

        // Switch line number only
        if(lineno){
          pos=editor->nextLine(0,lineno-1);
          editor->setCursorPos(pos);
          editor->setCenterLine(pos);
          }
        return 1;
        }
      }
    beep();
    }
    */
  return 1;
  }


// Close this window only
long Adie::onCmdClose(FXObject* sender,FXSelector,void*){
  TextWindow *window=(TextWindow*)sender;
  if(window->saveChanges()){
    windowlist.remove(window);
    return 0;
    }
  return 1;
  }


// Update title
long Adie::onUpdTitle(FXObject* sender,FXSelector,void*){
  TextWindow *window=(TextWindow*)sender;
  FXString title=FXFile::name(window->getFilename());
  FXString directory=FXFile::directory(window->getFilename());
  if(!directory.empty()) title+=" (" + directory + ")";
  if(window->isModified()) title+="*";
  window->setTitle(title);
  return 1;
  }


// Handle quit
long Adie::onCmdQuit(FXObject*,FXSelector,void*){
  // Loop over windows, see if OK to close
  FXTRACE((1,"quit\n"));
  exit(0);
  return 1;
  }


// User clicks on one of the window menus
long Adie::onCmdWindowSelect(FXObject*,FXSelector sel,void*){
  FXint which=FXSELID(sel)-ID_WINDOW_1;
  if(which<windowlist.no()){
    windowlist[which]->raise();
    windowlist[which]->setFocus();
    }
  return 1;
  }


// Update handler for window menus
long Adie::onUpdWindowSelect(FXObject *sender,FXSelector sel,void*){
  FXint which=FXSELID(sel)-ID_WINDOW_1;
  if(which<windowlist.no()){
    TextWindow *window=windowlist[which];
    FXString string;
    if(which<9)
      string.format("&%d %s",which+1,window->getTitle().text());
    else
      string.format("1&0 %s",window->getTitle().text());
    sender->handle(this,FXSEL(SEL_COMMAND,FXWindow::ID_SETSTRINGVALUE),(void*)&string);
    if(window==getFocusWindow())
      sender->handle(this,FXSEL(SEL_COMMAND,FXWindow::ID_CHECK),NULL);
    else
      sender->handle(this,FXSEL(SEL_COMMAND,FXWindow::ID_UNCHECK),NULL);
    sender->handle(this,FXSEL(SEL_COMMAND,FXWindow::ID_SHOW),NULL);
    }
  else{
    sender->handle(this,FXSEL(SEL_COMMAND,FXWindow::ID_HIDE),NULL);
    }
  return 1;
  }


// Clean up the mess
Adie::~Adie(){
  closeAllWindows();
  delete associations;
  delete bigicon;
  delete smallicon;
  delete newicon;
  delete reloadicon;
  delete openicon;
  delete saveicon;
  delete saveasicon;
  delete printicon;
  delete cuticon;
  delete copyicon;
  delete pasteicon;
  delete deleteicon;
  delete undoicon;
  delete redoicon;
  delete fontsicon;
  delete helpicon;
  delete quiticon;
  delete shiftlefticon;
  delete shiftrighticon;
  delete searchicon;
  delete searchnexticon;
  delete searchprevicon;
  delete bookseticon;
  delete booknexticon;
  delete bookprevicon;
  delete bookdelicon;
  }

